#include "pwrntfy.h"
#include <stdlib.h>

#define INT3 __asm int 3
#define DEBUGPRINT		KdPrint
#define ERRORPRINT(x)	KdPrint((x)); __asm int 3

//
// Handle to the shared event
//
HANDLE					SharedEventHandle = NULL;

//
// The actual shared event object
//
PKEVENT					SharedEvent	= NULL;


NTSTATUS
DriverEntry(
	PDRIVER_OBJECT pdrvo,
	PUNICODE_STRING uniRegistryPath
	) 
{
    NTSTATUS nts;
    PDEVICE_OBJECT pdevo;
    PDEVICE_EXTENSION pde;
	UNICODE_STRING uniDeviceName;
    UNICODE_STRING uniDosDeviceName;
	__int32 iFamily;

    UNREFERENCED_PARAMETER( uniRegistryPath );	// won't give a warning for not using param because macro makes it look used.
	DEBUGPRINT(("Entered DriverEntry\n"));		// a noop in release

	RtlInitUnicodeString( &uniDeviceName, UNI_DEVICE_NAME );

    nts = IoCreateDevice(
        pdrvo,
        sizeof( DEVICE_EXTENSION ),
        &uniDeviceName,
        FILE_DEVICE_UNKNOWN,
        0,
        FALSE, 
        &pdevo									
        );

	g_pde = pde = pdevo->DeviceExtension; // save pde globally for functions which don't get pde as param
	
	RtlInitUnicodeString( &uniDosDeviceName, UNI_DOS_DEVICE_NAME );

    nts = IoCreateSymbolicLink(				// this gives the driver a name user mode isn't blind to.
		(PUNICODE_STRING) &uniDosDeviceName,
		(PUNICODE_STRING) &uniDeviceName
    );

    pdrvo->DriverUnload = DriverUnload;			// register callbacks.
	pdrvo->MajorFunction[IRP_MJ_CREATE] =		
    pdrvo->MajorFunction[IRP_MJ_CLOSE] = DriverCreateClose;
    pdrvo->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DriverDeviceControl;
	pdrvo->Flags |= DO_BUFFERED_IO;				// yes, DO BUFFERED IO. sets up the system buffer.
	
	g_pvCallbackRegistration = NULL;

	return STATUS_SUCCESS;
}


VOID 
SetupPowerCallback(
	VOID
	)
{
	NTSTATUS nts;

	RtlInitUnicodeString(&g_UnicodeString, L"\\Callback\\PowerState"); // initialize string to proper location
	
	InitializeObjectAttributes(&g_oAttributes,			// initialize object to PowerState
							   &g_UnicodeString,
							   OBJ_CASE_INSENSITIVE,
							   NULL,
							   NULL);

	nts = ExCreateCallback(&g_pcObject,					// initialize callback object
						   &g_oAttributes,
				           TRUE,
					       TRUE);
	
	// try to setup callback func so on hibernate we can setup again
	if(nts == STATUS_SUCCESS)
		g_pvCallbackRegistration = ExRegisterCallback(g_pcObject,	// set func as callback func
													  DeviceDispatchPower,
													  NULL);
    return;
}


VOID
RemovePowerCallback(
    VOID
    )
{
	if ( g_pvCallbackRegistration )
	{
		ExUnregisterCallback(g_pvCallbackRegistration);  // unregister callback or we will get blue screen on power state change
	}

    return;
}


VOID
DriverUnload(
    IN PDRIVER_OBJECT pdrvo
    )
{
    PULONG Address;
	PDEVICE_OBJECT pdevo = pdrvo->DeviceObject;
    PDEVICE_EXTENSION pde = pdevo->DeviceExtension;
	UNICODE_STRING uniDosDeviceName;
	int i;

	DEBUGPRINT( ( "Entering DriverUnload\n" ) );

	if (SharedEventHandle != NULL) ZwClose(SharedEventHandle);
    
	KeClearEvent( SharedEvent );
    RemovePowerCallback();                  
	
	RtlInitUnicodeString( &uniDosDeviceName, UNI_DOS_DEVICE_NAME );  // take us out of windows services
    IoDeleteSymbolicLink( &uniDosDeviceName );
    IoDeleteDevice( pdevo );
	
	DEBUGPRINT(("Leaving DriverUnload\n"));

    return;
}


NTSTATUS
DriverCreateClose(
    IN PDEVICE_OBJECT pdevo,
    IN PIRP pirp
    )
{
	NTSTATUS nts = STATUS_SUCCESS;
	PIO_STACK_LOCATION pirpstk;                     
    PDEVICE_EXTENSION pde = pdevo->DeviceExtension;

	pirpstk = IoGetCurrentIrpStackLocation(pirp);

    pirp->IoStatus.Status = STATUS_SUCCESS;
    pirp->IoStatus.Information = 0;
    IoCompleteRequest(pirp, IO_NO_INCREMENT);
	
    return nts;
}


NTSTATUS
DriverDeviceControl(	
    IN PDEVICE_OBJECT pdevo,
    IN PIRP pirp
    )
{
    NTSTATUS nts = STATUS_SUCCESS;
	PIO_STACK_LOCATION pirpstk;
    PDEVICE_EXTENSION pde = pdevo->DeviceExtension;
	int i;
	UNICODE_STRING	   EventName;

	DEBUGPRINT(("Entering DriverDeviceControl\n"));

	pirpstk = IoGetCurrentIrpStackLocation( pirp );
    //
    // Set the status field to zero for completion
    //
    pirp->IoStatus.Information = 0;
    
    switch (pirpstk->Parameters.DeviceIoControl.IoControlCode) 
    {
	case IOCTL_PWRNTFY_EVENT:
		//
		// When you create a named event in user mode \BaseNamedObjects\ is 
		// prepended to the name specified in the call to CreateEvent
		//
		RtlInitUnicodeString(&EventName, L"\\BaseNamedObjects\\SharedEvent");

		//
		// Get a pointer to a KEVENT object that is associated with this
		// named event
		//
		SharedEvent = IoCreateNotificationEvent(&EventName, &SharedEventHandle);

		if (SharedEvent != NULL) 
			nts = STATUS_SUCCESS;
        else 
			nts = STATUS_UNSUCCESSFUL;
	
		break;

	case IOCTL_PWRNTFY_EVENT_START:
		SetupPowerCallback();
		nts = STATUS_SUCCESS;
		break;

    default:
		nts = STATUS_INVALID_PARAMETER;
        break;
    }

	pirp->IoStatus.Status = nts;
	IoCompleteRequest( pirp, IO_NO_INCREMENT);

	//DEBUGPRINT(("Leaving DriverDeviceControl\n"));

    return nts;
}


VOID
DeviceDispatchPower (
    IN PVOID CallbackContext,
    IN PVOID Argument1,
    IN PVOID Argument2
    )
{
	//
	// Set the event so that the user mode application's wait will
	// be satisfied. Note that the third parameter, Wait, must be
	// set to FALSE since we're running at DISPATCH_LEVEL
	//

	KeSetEvent(SharedEvent, 0, FALSE);

    return;
}



